﻿(* # F# Extensions to System.Net.Http

## License

Author: Ryan Riley <ryan.riley@panesofglass.org>
Copyright (c) 2011-2012, Ryan Riley.

Licensed under the Apache License, Version 2.0.
See LICENSE.txt for details.
*)
namespace System.Net.Http

open System.Net.Http
open System.Net.Http.Formatting
open System.Net.Http.Headers
open System.Threading.Tasks

// ## Define the web application interface

(*
One may define a web application interface using a large variety of signatures. Indeed, if you search the web, you're likely to find a large number of approaches. When starting with `Frank`, I wanted to try to find a way to define an HTTP application using pure functions and function composition. The closest I found was the following:

        type HttpApplication = HttpRequestMessage -> Async<HttpResponseMessage>

Alas, this approach works only so well. HTTP is a rich communication specification. The simplicity and elegance of a purely functional approach quickly loses the ability to communicate back options to the client. For instance, given the above, how do you return a meaningful `405 Method Not Allowed` response? The HTTP specification requires that you list the allowed methods, but if you merge all the logic for selecting an application into the functions, there is no easy way to recall all the allowed methods, short of trying them all. You could require that the developer add the list of used methods, but that, too, misses the point that the application should be collecting this and helping the developer by taking care of all of the nuts and bolts items.

The next approach I tried involved using a tuple of a list of allowed HTTP methods and the application handler, which used the merged function approach described above for actually executing the application. However, once again, there are limitations. This structure accurately represents a resource, but it does not allow for multiple resources to coexist side-by-side. Another tuple of uri pattern matching expressions could wrap a list of these method * handler tuples, but at this point I realized I would be better served by using real types and thus arrived at the signatures below.

You'll see the signatures above are still mostly present, though they have been changed to better fit the signatures below.
*)

// `HttpApplication` defines the contract for processing any request.
// An application takes an `HttpRequestMessage` and returns an `HttpRequestHandler` asynchronously.
type HttpApplication = HttpRequestMessage -> Async<HttpResponseMessage>

type EmptyContent() =
    inherit HttpContent()
    override x.SerializeToStreamAsync(stream, context) =
        let tcs = new TaskCompletionSource<_>(TaskCreationOptions.None)
        tcs.SetResult(())
        tcs.Task :> Task
    override x.TryComputeLength(length) =
        length <- 0L
        true
    override x.Equals(other) =
        other.GetType() = typeof<EmptyContent>
    override x.GetHashCode() = hash x

type AsyncHandler =
    inherit DelegatingHandler
    val AsyncSend : HttpRequestMessage -> Async<HttpResponseMessage>
    new (f, inner) = { inherit DelegatingHandler(inner); AsyncSend = f }
    new (f) = { inherit DelegatingHandler(); AsyncSend = f }
    override x.SendAsync(request, cancellationToken) =
        Async.StartAsTask(x.AsyncSend request, cancellationToken = cancellationToken)

[<AutoOpen>]
module Extensions =
    open System
    open System.Net
    open System.Net.Http
    open System.Net.Http.Headers

    let private emptyContent = new EmptyContent() :> HttpContent

    type HttpContent with
        static member Empty = emptyContent
        member x.AsyncReadAs<'a>() = Async.AwaitTask <| x.ReadAsAsync<'a>()
        member x.AsyncReadAs<'a>(formatters:Formatting.MediaTypeFormatter seq) = Async.AwaitTask <| x.ReadAsAsync<'a>(formatters)
        member x.AsyncReadAs(type': Type) = Async.AwaitTask <| x.ReadAsAsync(type')
        member x.AsyncReadAs(type': Type, formatters: Formatting.MediaTypeFormatter seq) = Async.AwaitTask <| x.ReadAsAsync(type', formatters)
        member x.AsyncReadAsByteArray() = Async.AwaitTask <| x.ReadAsByteArrayAsync()
        member x.AsyncReadAsHttpRequestMessage() = Async.AwaitTask <| x.ReadAsHttpRequestMessageAsync()
        member x.AsyncReadAsHttpResponseMessage() = Async.AwaitTask <| x.ReadAsHttpResponseMessageAsync()
        member x.AsyncReadAsMultipart() = Async.AwaitTask <| x.ReadAsMultipartAsync()
        member x.AsyncReadAsMultipart(streamProvider) = Async.AwaitTask <| x.ReadAsMultipartAsync(streamProvider)
        member x.AsyncReadAsMultipart(streamProvider, bufferSize:int) = Async.AwaitTask <| x.ReadAsMultipartAsync(streamProvider, bufferSize)
        member x.AsyncReadAsStream() = Async.AwaitTask <| x.ReadAsStreamAsync()
        member x.AsyncReadAsString() = Async.AwaitTask <| x.ReadAsStringAsync()
